@@ -1,9 +1,9 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from trytond.model import ModelView, ModelSQL, DeactivableMixin, fields, Unique
-from trytond.wizard import Wizard, StateView, Button
-from trytond.pyson import Eval
+from trytond.model import DeactivableMixin, ModelSQL, ModelView, fields
 from trytond.pool import Pool
+from trytond.pyson import Eval
+from trytond.wizard import Button, StateView, Wizard
 
 
 class BOM(DeactivableMixin, ModelSQL, ModelView):
@@ -16,20 +16,20 @@ class BOM(DeactivableMixin, ModelSQL, ModelView):
     output_products = fields.Many2Many('production.bom.output',
         'bom', 'product', 'Output Products')
 
-    def compute_factor(self, product, quantity, uom):
+    def compute_factor(self, product, quantity, unit):
         '''
         Compute factor for an output product
         '''
         Uom = Pool().get('product.uom')
+        output_quantity = 0
         for output in self.outputs:
             if output.product == product:
-                if not output.quantity:
-                    return 0.0
-                quantity = Uom.compute_qty(uom, quantity,
-                    output.uom, round=False)
-                return quantity / output.quantity
+                output_quantity += Uom.compute_qty(
+                    output.unit, output.quantity, unit, round=False)
+        if output_quantity:
+            return quantity / output_quantity
         else:
-            return 0.0
+            return 0
 
     @classmethod
     def copy(cls, records, default=None):
@@ -45,60 +45,60 @@ class BOMInput(ModelSQL, ModelView):
     "Bill of Material Input"
     __name__ = 'production.bom.input'
 
-    bom = fields.Many2One('production.bom', 'BOM', required=True,
-        select=1, ondelete='CASCADE')
+    bom = fields.Many2One(
+        'production.bom', "BOM", required=True, ondelete='CASCADE')
     product = fields.Many2One('product.product', 'Product', required=True)
     uom_category = fields.Function(fields.Many2One(
         'product.uom.category', 'Uom Category'), 'on_change_with_uom_category')
-    uom = fields.Many2One('product.uom', 'Uom', required=True,
+    unit = fields.Many2One(
+        'product.uom', "Unit", required=True,
         domain=[
             ('category', '=', Eval('uom_category')),
-        ], depends=['uom_category'])
-    unit_digits = fields.Function(fields.Integer('Unit Digits'),
-        'on_change_with_unit_digits')
-    quantity = fields.Float('Quantity', required=True,
+            ])
+    quantity = fields.Float(
+        "Quantity", digits='unit', required=True,
         domain=['OR',
             ('quantity', '>=', 0),
             ('quantity', '=', None),
-            ],
-        digits=(16, Eval('unit_digits', 2)),
-        depends=['unit_digits'])
+            ])
 
     @classmethod
     def __setup__(cls):
         super(BOMInput, cls).__setup__()
         cls.product.domain = [('type', 'in', cls.get_product_types())]
         cls.__access__.add('bom')
-        t = cls.__table__()
-        cls._sql_constraints = [
-            ('product_bom_uniq', Unique(t, t.product, t.bom),
-                'production.msg_product_bom_unique'),
-            ]
+
+    @classmethod
+    def __register__(cls, module):
+        table_h = cls.__table_handler__(module)
+
+        # Migration from 6.8: rename uom to unit
+        if (table_h.column_exist('uom')
+                and not table_h.column_exist('unit')):
+            table_h.column_rename('uom', 'unit')
+
+        super().__register__(module)
+        table_h = cls.__table_handler__(module)
+
+        # Migration from 6.0: remove unique constraint
+        table_h.drop_constraint('product_bom_uniq')
 
     @classmethod
     def get_product_types(cls):
         return ['goods', 'assets']
 
-    @fields.depends('product', 'uom')
+    @fields.depends('product', 'unit')
     def on_change_product(self):
         if self.product:
             category = self.product.default_uom.category
-            if not self.uom or self.uom.category != category:
-                self.uom = self.product.default_uom
-                self.unit_digits = self.product.default_uom.digits
+            if not self.unit or self.unit.category != category:
+                self.unit = self.product.default_uom
         else:
-            self.uom = None
+            self.unit = None
 
     @fields.depends('product')
     def on_change_with_uom_category(self, name=None):
-        if self.product:
-            return self.product.default_uom.category.id
-
-    @fields.depends('uom')
-    def on_change_with_unit_digits(self, name=None):
-        if self.uom:
-            return self.uom.digits
-        return 2
+        return self.product.default_uom.category if self.product else None
 
     def get_rec_name(self, name):
         return self.product.rec_name
@@ -120,7 +120,7 @@ class BOMInput(ModelSQL, ModelView):
         self.product.check_bom_recursion()
 
     def compute_quantity(self, factor):
-        return self.uom.ceil(self.quantity * factor)
+        return self.unit.ceil(self.quantity * factor)
 
 
 class BOMOutput(BOMInput):
@@ -129,7 +129,7 @@ class BOMOutput(BOMInput):
     _table = 'production_bom_output'  # Needed to override BOMInput._table
 
     def compute_quantity(self, factor):
-        return self.uom.floor(self.quantity * factor)
+        return self.unit.floor(self.quantity * factor)
 
     @classmethod
     def delete(cls, outputs):
@@ -140,20 +140,32 @@ class BOMOutput(BOMInput):
         # Validate that output_products domain on bom is still valid
         ProductBOM._validate(bom_products, ['bom'])
 
+    @classmethod
+    def write(cls, *args):
+        pool = Pool()
+        ProductBOM = pool.get('product.product-production.bom')
+        actions = iter(args)
+        bom_products = []
+        for outputs, values in zip(actions, actions):
+            if 'product' in values:
+                bom_products.extend(
+                    [b for o in outputs for b in o.product.boms])
+        super().write(*args)
+        # Validate that output_products domain on bom is still valid
+        ProductBOM._validate(bom_products, ['bom'])
+
 
 class BOMTree(ModelView):
     'BOM Tree'
     __name__ = 'production.bom.tree'
 
     product = fields.Many2One('product.product', 'Product')
-    quantity = fields.Float('Quantity', digits=(16, Eval('unit_digits', 2)),
-        depends=['unit_digits'])
-    uom = fields.Many2One('product.uom', 'Uom')
-    unit_digits = fields.Integer('Unit Digits')
+    quantity = fields.Float('Quantity', digits='unit')
+    unit = fields.Many2One('product.uom', "Unit")
     childs = fields.One2Many('production.bom.tree', None, 'Childs')
 
     @classmethod
-    def tree(cls, product, quantity, uom, bom=None):
+    def tree(cls, product, quantity, unit, bom=None):
         Input = Pool().get('production.bom.input')
 
         result = []
@@ -162,21 +174,20 @@ class BOMTree(ModelView):
                 return result
             bom = product.boms[0].bom
 
-        factor = bom.compute_factor(product, quantity, uom)
+        factor = bom.compute_factor(product, quantity, unit)
         for input_ in bom.inputs:
             quantity = Input.compute_quantity(input_, factor)
-            childs = cls.tree(input_.product, quantity, input_.uom)
+            childs = cls.tree(input_.product, quantity, input_.unit)
             values = {
                 'product': input_.product.id,
                 'product.': {
                     'rec_name': input_.product.rec_name,
                     },
                 'quantity': quantity,
-                'uom': input_.uom.id,
-                'uom.': {
-                    'rec_name': input_.uom.rec_name,
+                'unit': input_.unit.id,
+                'unit.': {
+                    'rec_name': input_.unit.rec_name,
                     },
-                'unit_digits': input_.uom.digits,
                 'childs': childs,
             }
             result.append(values)
@@ -187,27 +198,20 @@ class OpenBOMTreeStart(ModelView):
     'Open BOM Tree'
     __name__ = 'production.bom.tree.open.start'
 
-    quantity = fields.Float('Quantity', required=True,
-        digits=(16, Eval('unit_digits', 2)), depends=['unit_digits'])
-    uom = fields.Many2One('product.uom', 'Unit', required=True,
+    quantity = fields.Float('Quantity', digits='unit', required=True)
+    unit = fields.Many2One(
+        'product.uom', "Unit", required=True,
         domain=[
             ('category', '=', Eval('category')),
-        ], depends=['category'])
-    unit_digits = fields.Integer('Unit Digits', readonly=True)
+            ])
     category = fields.Many2One('product.uom.category', 'Category',
         readonly=True)
     bom = fields.Many2One('product.product-production.bom',
         'BOM', required=True, domain=[
             ('product', '=', Eval('product')),
-        ], depends=['product'])
+            ])
     product = fields.Many2One('product.product', 'Product', readonly=True)
 
-    @fields.depends('uom')
-    def on_change_with_unit_digits(self):
-        if self.uom:
-            return self.uom.digits
-        return 2
-
 
 class OpenBOMTreeTree(ModelView):
     'Open BOM Tree'
@@ -217,22 +221,21 @@ class OpenBOMTreeTree(ModelView):
         readonly=True)
 
     @classmethod
-    def tree(cls, bom, product, quantity, uom):
+    def tree(cls, bom, product, quantity, unit):
         pool = Pool()
         Tree = pool.get('production.bom.tree')
 
-        childs = Tree.tree(product, quantity, uom, bom=bom)
+        childs = Tree.tree(product, quantity, unit, bom=bom)
         bom_tree = [{
                 'product': product.id,
                 'product.': {
                     'rec_name': product.rec_name,
                     },
                 'quantity': quantity,
-                'uom': uom.id,
-                'uom.': {
-                    'rec_name': uom.rec_name,
+                'unit': unit.id,
+                'unit.': {
+                    'rec_name': unit.rec_name,
                     },
-                'unit_digits': uom.digits,
                 'childs': childs,
                 }]
         return {
@@ -259,12 +262,10 @@ class OpenBOMTree(Wizard):
         defaults = {}
         product = self.record
         defaults['category'] = product.default_uom.category.id
-        if getattr(self.start, 'uom', None):
-            defaults['uom'] = self.start.uom.id
-            defaults['unit_digits'] = self.start.unit_digits
+        if getattr(self.start, 'unit', None):
+            defaults['unit'] = self.start.unit.id
         else:
-            defaults['uom'] = product.default_uom.id
-            defaults['unit_digits'] = product.default_uom.digits
+            defaults['unit'] = product.default_uom.id
         defaults['product'] = product.id
         if getattr(self.start, 'bom', None):
             defaults['bom'] = self.start.bom.id
@@ -277,4 +278,4 @@ class OpenBOMTree(Wizard):
         pool = Pool()
         BomTree = pool.get('production.bom.tree.open.tree')
         return BomTree.tree(self.start.bom.bom, self.start.product,
-            self.start.quantity, self.start.uom)
+            self.start.quantity, self.start.unit)
