@@ -1,11 +1,11 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 from trytond.i18n import gettext
-from trytond.model import ModelView, ModelSQL, MatchMixin, fields, \
-    sequence_ordered
+from trytond.model import (
+    MatchMixin, ModelSQL, ModelView, fields, sequence_ordered)
 from trytond.model.exceptions import RecursionError
-from trytond.pyson import Eval, Get, If, Bool
 from trytond.pool import PoolMeta
+from trytond.pyson import Bool, Eval, Get, If, TimeDelta
 
 
 class Template(metaclass=PoolMeta):
@@ -18,7 +18,6 @@ class Template(metaclass=PoolMeta):
         cls.producible.states = {
             'invisible': ~Eval('type').in_(cls.get_producible_types()),
             }
-        cls.producible.depends = ['type']
 
     @classmethod
     def get_producible_types(cls):
@@ -39,14 +38,12 @@ class Product(metaclass=PoolMeta):
         'BOMs', order=[('sequence', 'ASC'), ('id', 'ASC')],
         states={
             'invisible': ~Eval('producible')
-            },
-        depends=['producible'])
-    lead_times = fields.One2Many('production.lead_time',
+            })
+    production_lead_times = fields.One2Many('production.lead_time',
         'product', 'Lead Times', order=[('sequence', 'ASC'), ('id', 'ASC')],
         states={
             'invisible': ~Eval('producible'),
-            },
-        depends=['producible'])
+            })
 
     @classmethod
     def validate(cls, products):
@@ -76,7 +73,7 @@ class Product(metaclass=PoolMeta):
         else:
             default = default.copy()
         default.setdefault('boms', None)
-        default.setdefault('lead_times', None)
+        default.setdefault('production_lead_times', None)
         return super(Product, cls).copy(products, default=default)
 
 
@@ -84,17 +81,18 @@ class ProductBom(sequence_ordered(), ModelSQL, ModelView):
     'Product - BOM'
     __name__ = 'product.product-production.bom'
 
-    product = fields.Many2One('product.product', 'Product',
-        ondelete='CASCADE', select=1, required=True,
+    product = fields.Many2One(
+        'product.product', "Product", ondelete='CASCADE', required=True,
         domain=[
             ('producible', '=', True),
             ])
-    bom = fields.Many2One('production.bom', 'BOM', ondelete='CASCADE',
-        select=1, required=True, domain=[
+    bom = fields.Many2One(
+        'production.bom', "BOM", ondelete='CASCADE', required=True,
+        domain=[
             ('output_products', '=', If(Bool(Eval('product')),
                     Eval('product', 0),
                     Get(Eval('_parent_product', {}), 'id', 0))),
-            ], depends=['product'])
+            ])
 
     def get_rec_name(self, name):
         return self.bom.rec_name
@@ -108,8 +106,8 @@ class ProductionLeadTime(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
     'Production Lead Time'
     __name__ = 'production.lead_time'
 
-    product = fields.Many2One('product.product', 'Product',
-        ondelete='CASCADE', select=True, required=True,
+    product = fields.Many2One(
+        'product.product', "Product", ondelete='CASCADE', required=True,
         domain=[
             ('producible', '=', True),
             ])
@@ -118,9 +116,13 @@ class ProductionLeadTime(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
             ('output_products', '=', If(Bool(Eval('product')),
                     Eval('product', -1),
                     Get(Eval('_parent_product', {}), 'id', 0))),
-            ],
-        depends=['product'])
-    lead_time = fields.TimeDelta('Lead Time')
+            ])
+    lead_time = fields.TimeDelta(
+        "Lead Time",
+        domain=['OR',
+            ('lead_time', '=', None),
+            ('lead_time', '>=', TimeDelta()),
+            ])
 
     @classmethod
     def __setup__(cls):
