@@ -3,61 +3,57 @@
 from collections import defaultdict
 from datetime import timedelta
 from decimal import Decimal
-from itertools import chain
+from itertools import chain, groupby
 
-from sql import Null
 from sql.conditionals import Coalesce
+from sql.functions import CharLength
 
-from trytond.model import ModelView, ModelSQL, Workflow, fields, dualmethod
-from trytond.pyson import Eval, Bool, If
-from trytond.pool import Pool
-from trytond.transaction import Transaction
-
+from trytond.i18n import gettext
+from trytond.model import (
+    Index, ModelSQL, ModelView, Workflow, dualmethod, fields)
 from trytond.modules.company.model import employee_field, set_employee
 from trytond.modules.product import price_digits, round_price
 from trytond.modules.stock.shipment import ShipmentAssignMixin
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval, If
+from trytond.transaction import Transaction
 
-BOM_CHANGES = ['bom', 'product', 'quantity', 'uom', 'warehouse', 'location',
-    'company', 'inputs', 'outputs']
+from .exceptions import CostWarning
 
 
 class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
     "Production"
     __name__ = 'production'
+    _rec_name = 'number'
     _assign_moves_field = 'inputs'
 
-    number = fields.Char('Number', select=True, readonly=True)
-    reference = fields.Char('Reference', select=1,
+    number = fields.Char("Number", readonly=True)
+    reference = fields.Char(
+        "Reference",
         states={
             'readonly': ~Eval('state').in_(['request', 'draft']),
-            },
-        depends=['state'])
+            })
     planned_date = fields.Date('Planned Date',
         states={
-            'readonly': ~Eval('state').in_(['request', 'draft']),
-            },
-        depends=['state'])
+            'readonly': Eval('state').in_(['cancelled', 'done']),
+            })
     effective_date = fields.Date('Effective Date',
         states={
             'readonly': Eval('state').in_(['cancelled', 'done']),
-            },
-        depends=['state'])
+            })
     planned_start_date = fields.Date('Planned Start Date',
         states={
             'readonly': ~Eval('state').in_(['request', 'draft']),
             'required': Bool(Eval('planned_date')),
-            },
-        depends=['state', 'planned_date'])
+            })
     effective_start_date = fields.Date('Effective Start Date',
         states={
             'readonly': Eval('state').in_(['cancelled', 'running', 'done']),
-            },
-        depends=['state'])
+            })
     company = fields.Many2One('company.company', 'Company', required=True,
         states={
             'readonly': ~Eval('state').in_(['request', 'draft']),
-            },
-        depends=['state'])
+            })
     warehouse = fields.Many2One('stock.location', 'Warehouse', required=True,
         domain=[
             ('type', '=', 'warehouse'),
@@ -65,8 +61,7 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
         states={
             'readonly': (~Eval('state').in_(['request', 'draft'])
                 | Eval('inputs', [-1]) | Eval('outputs', [-1])),
-            },
-        depends=['state'])
+            })
     location = fields.Many2One('stock.location', 'Location', required=True,
         domain=[
             ('type', '=', 'production'),
@@ -74,8 +69,7 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
         states={
             'readonly': (~Eval('state').in_(['request', 'draft'])
                 | Eval('inputs', [-1]) | Eval('outputs', [-1])),
-            },
-        depends=['state'])
+            })
     product = fields.Many2One('product.product', 'Product',
         domain=[
             ('producible', '=', True),
@@ -86,7 +80,7 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
     bom = fields.Many2One('production.bom', 'BOM',
         domain=[
             ('output_products', '=', Eval('product', 0)),
@@ -95,12 +89,13 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
             'readonly': (~Eval('state').in_(['request', 'draft'])
                 | ~Eval('warehouse', 0) | ~Eval('location', 0)),
             'invisible': ~Eval('product'),
-            },
-        depends=['product'])
+            })
     uom_category = fields.Function(fields.Many2One(
-            'product.uom.category', 'Uom Category'),
+            'product.uom.category', "UoM Category",
+            help="The category of Unit of Measure."),
         'on_change_with_uom_category')
-    uom = fields.Many2One('product.uom', 'Uom',
+    unit = fields.Many2One(
+        'product.uom', "Unit",
         domain=[
             ('category', '=', Eval('uom_category')),
             ],
@@ -108,47 +103,41 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
             'readonly': ~Eval('state').in_(['request', 'draft']),
             'required': Bool(Eval('bom')),
             'invisible': ~Eval('product'),
-            },
-        depends=['uom_category'])
-    unit_digits = fields.Function(fields.Integer('Unit Digits'),
-        'on_change_with_unit_digits')
-    quantity = fields.Float('Quantity',
-        digits=(16, Eval('unit_digits', 2)),
+            })
+    quantity = fields.Float(
+        "Quantity", digits='unit',
         states={
             'readonly': ~Eval('state').in_(['request', 'draft']),
             'required': Bool(Eval('bom')),
             'invisible': ~Eval('product'),
-            },
-        depends=['unit_digits'])
+            })
     cost = fields.Function(fields.Numeric('Cost', digits=price_digits,
             readonly=True), 'get_cost')
     inputs = fields.One2Many('stock.move', 'production_input', 'Inputs',
         domain=[
             ('shipment', '=', None),
-            ('from_location', 'child_of', [Eval('warehouse')], 'parent'),
-            ('to_location', '=', Eval('location')),
+            ('from_location', 'child_of', [Eval('warehouse', -1)], 'parent'),
+            ('to_location', '=', Eval('location', -1)),
             ('company', '=', Eval('company', -1)),
             ],
         states={
             'readonly': (~Eval('state').in_(['request', 'draft', 'waiting'])
                 | ~Eval('warehouse') | ~Eval('location')),
-            },
-        depends=['warehouse', 'location', 'company'])
+            })
     outputs = fields.One2Many('stock.move', 'production_output', 'Outputs',
         domain=[
             ('shipment', '=', None),
-            ('from_location', '=', Eval('location')),
+            ('from_location', '=', Eval('location', -1)),
             ['OR',
-                ('to_location', 'child_of', [Eval('warehouse')], 'parent'),
-                ('to_location.waste_warehouses', '=', Eval('warehouse')),
+                ('to_location', 'child_of', [Eval('warehouse', -1)], 'parent'),
+                ('to_location.waste_warehouses', '=', Eval('warehouse', -1)),
                 ],
             ('company', '=', Eval('company', -1)),
             ],
         states={
             'readonly': (Eval('state').in_(['done', 'cancelled'])
                 | ~Eval('warehouse') | ~Eval('location')),
-            },
-        depends=['warehouse', 'location', 'company'])
+            })
 
     assigned_by = employee_field("Assigned By")
     run_by = employee_field("Run By")
@@ -161,17 +150,28 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
             ('running', 'Running'),
             ('done', 'Done'),
             ('cancelled', 'Cancelled'),
-            ], 'State', readonly=True)
+            ], 'State', readonly=True, sort=False)
     origin = fields.Reference(
-        "Origin", selection='get_origin', select=True,
+        "Origin", selection='get_origin',
         states={
             'readonly': ~Eval('state').in_(['request', 'draft']),
-            },
-        depends=['state'])
+            })
 
     @classmethod
     def __setup__(cls):
+        cls.number.search_unaccented = False
+        cls.reference.search_unaccented = False
         super(Production, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(t, (t.reference, Index.Similarity())),
+                Index(
+                    t,
+                    (t.state, Index.Equality()),
+                    where=t.state.in_([
+                            'request', 'draft', 'waiting', 'assigned',
+                            'running'])),
+                })
         cls._order = [
             ('effective_date', 'ASC NULLS LAST'),
             ('id', 'ASC'),
@@ -263,27 +263,27 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
     @classmethod
     def __register__(cls, module_name):
         table_h = cls.__table_handler__(module_name)
-        table = cls.__table__()
 
-        # Migration from 3.8: rename code into number
-        if table_h.column_exist('code'):
-            table_h.column_rename('code', 'number')
+        # Migration from 6.8: rename uom to unit
+        if (table_h.column_exist('uom')
+                and not table_h.column_exist('unit')):
+            table_h.column_rename('uom', 'unit')
 
         super(Production, cls).__register__(module_name)
 
-        # Migration from 4.0: fill planned_start_date
+        table = cls.__table__()
         cursor = Transaction().connection.cursor()
-        cursor.execute(*table.update(
-                [table.planned_start_date],
-                [table.planned_date],
-                where=(table.planned_start_date == Null)
-                & (table.planned_date != Null)))
 
         # Migration from 5.6: rename state cancel to cancelled
         cursor.execute(*table.update(
                 [table.state], ['cancelled'],
                 where=table.state == 'cancel'))
 
+    @classmethod
+    def order_number(cls, tables):
+        table, _ = tables[None]
+        return [CharLength(table.number), table.number]
+
     @classmethod
     def order_effective_date(cls, tables):
         table, _ = tables[None]
@@ -314,20 +314,27 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
 
     @fields.depends('product', 'bom')
     def compute_lead_time(self, pattern=None):
-        if pattern is None:
-            pattern = {}
+        pattern = pattern.copy() if pattern is not None else {}
         if self.product:
             pattern.setdefault('bom', self.bom.id if self.bom else None)
-            for line in self.product.lead_times:
+            for line in self.product.production_lead_times:
                 if line.match(pattern):
                     return line.lead_time or timedelta()
         return timedelta()
 
-    @fields.depends('planned_date', methods=['compute_lead_time'])
-    def on_change_with_planned_start_date(self, pattern=None):
-        if self.planned_date and self.product:
-            return self.planned_date - self.compute_lead_time()
-        return self.planned_date
+    @fields.depends(
+        'planned_date', 'state', 'product', methods=['compute_lead_time'])
+    def set_planned_start_date(self):
+        if self.state in {'request', 'draft'}:
+            if self.planned_date and self.product:
+                self.planned_start_date = (
+                    self.planned_date - self.compute_lead_time())
+            else:
+                self.planned_start_date = self.planned_date
+
+    @fields.depends(methods=['set_planned_start_date'])
+    def on_change_planned_date(self):
+        self.set_planned_start_date()
 
     @fields.depends(
         'planned_date', 'planned_start_date', methods=['compute_lead_time'])
@@ -350,59 +357,63 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
         models = cls._get_origin()
         return [(None, '')] + [(m, get_name(m)) for m in models]
 
-    def _move(self, from_location, to_location, company, product, uom,
-            quantity):
-        Move = Pool().get('stock.move')
+    @fields.depends(
+        'company', 'location', methods=['picking_location', 'output_location'])
+    def _move(self, type, product, unit, quantity):
+        pool = Pool()
+        Move = pool.get('stock.move')
+        assert type in {'input', 'output'}
         move = Move(
             product=product,
-            uom=uom,
+            unit=unit,
             quantity=quantity,
-            from_location=from_location,
-            to_location=to_location,
-            company=company,
-            currency=company.currency if company else None,
-            state='draft',
-            )
+            company=self.company,
+            currency=self.company.currency if self.company else None)
+        if type == 'input':
+            move.from_location = self.picking_location
+            move.to_location = self.location
+            move.production_input = self
+        else:
+            move.from_location = self.location
+            move.to_location = self.output_location
+            move.production_output = self
         return move
 
-    def _explode_move_values(self, from_location, to_location, company,
-            bom_io, quantity):
-        move = self._move(from_location, to_location, company,
-            bom_io.product, bom_io.uom, quantity)
-        move.from_location = from_location.id if from_location else None
-        move.to_location = to_location.id if to_location else None
+    @fields.depends(methods=['_move'])
+    def _explode_move_values(self, type, bom_io, quantity):
+        move = self._move(type, bom_io.product, bom_io.unit, quantity)
         move.unit_price_required = move.on_change_with_unit_price_required()
         return move
 
-    @fields.depends(*BOM_CHANGES)
+    @fields.depends(
+        'bom', 'product', 'unit', 'quantity', 'company', 'inputs', 'outputs',
+        methods=['_explode_move_values'])
     def explode_bom(self):
         pool = Pool()
         Uom = pool.get('product.uom')
-        if not (self.bom and self.product and self.uom):
+        if not (self.bom and self.product and self.unit):
             return
 
-        factor = self.bom.compute_factor(self.product, self.quantity or 0,
-            self.uom)
+        factor = self.bom.compute_factor(
+            self.product, self.quantity or 0, self.unit)
         inputs = []
         for input_ in self.bom.inputs:
             quantity = input_.compute_quantity(factor)
-            move = self._explode_move_values(
-                self.picking_location, self.location, self.company,
-                input_, quantity)
+            move = self._explode_move_values('input', input_, quantity)
             if move:
                 inputs.append(move)
-                quantity = Uom.compute_qty(input_.uom, quantity,
-                    input_.product.default_uom, round=False)
+                quantity = Uom.compute_qty(
+                    input_.unit, quantity, input_.product.default_uom,
+                    round=False)
         self.inputs = inputs
 
         outputs = []
         for output in self.bom.outputs:
             quantity = output.compute_quantity(factor)
-            move = self._explode_move_values(
-                self.location, self.output_location, self.company, output,
-                quantity)
+            move = self._explode_move_values('output', output, quantity)
             if move:
                 move.unit_price = Decimal(0)
+                move.currency = self.company.currency
                 outputs.append(move)
         self.outputs = outputs
 
@@ -412,36 +423,31 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
         if self.warehouse:
             self.location = self.warehouse.production_location
 
-    @fields.depends('product', 'uom', methods=['explode_bom'])
+    @fields.depends(
+        'product', 'unit', methods=['explode_bom', 'set_planned_start_date'])
     def on_change_product(self):
         if self.product:
             category = self.product.default_uom.category
-            if not self.uom or self.uom.category != category:
-                self.uom = self.product.default_uom
-                self.unit_digits = self.product.default_uom.digits
+            if not self.unit or self.unit.category != category:
+                self.unit = self.product.default_uom
         else:
             self.bom = None
-            self.uom = None
-            self.unit_digits = 2
+            self.unit = None
         self.explode_bom()
+        self.set_planned_start_date()
 
     @fields.depends('product')
     def on_change_with_uom_category(self, name=None):
-        if self.product:
-            return self.product.default_uom.category.id
+        return self.product.default_uom.category if self.product else None
 
-    @fields.depends('uom')
-    def on_change_with_unit_digits(self, name=None):
-        if self.uom:
-            return self.uom.digits
-        return 2
-
-    @fields.depends(methods=['explode_bom'])
+    @fields.depends(methods=['explode_bom', 'set_planned_start_date'])
     def on_change_bom(self):
         self.explode_bom()
+        # Product's production lead time depends on bom
+        self.set_planned_start_date()
 
     @fields.depends(methods=['explode_bom'])
-    def on_change_uom(self):
+    def on_change_unit(self):
         self.explode_bom()
 
     @fields.depends(methods=['explode_bom'])
@@ -457,10 +463,7 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
         for input_ in self.inputs:
             if input_.state == 'cancelled':
                 continue
-            if input_.cost_price is not None:
-                cost_price = input_.cost_price
-            else:
-                cost_price = input_.product.cost_price
+            cost_price = input_.get_cost_price()
             cost += (Decimal(str(input_.internal_quantity)) * cost_price)
         return round_price(cost)
 
@@ -474,13 +477,13 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
 
         for input_ in self.inputs:
             if (input_.product is None
-                    or input_.uom is None
+                    or input_.unit is None
                     or input_.quantity is None
                     or input_.state == 'cancelled'):
                 continue
             product = input_.product
-            quantity = Uom.compute_qty(input_.uom, input_.quantity,
-                product.default_uom)
+            quantity = Uom.compute_qty(
+                input_.unit, input_.quantity, product.default_uom)
             cost += Decimal(str(quantity)) * product.cost_price
         return cost
 
@@ -490,51 +493,39 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
         Move = pool.get('stock.move')
         to_save = []
         for production in productions:
-            location = production.location
-            company = production.company
-
             if not production.bom:
                 if production.product:
                     move = production._move(
-                        location, production.output_location, company,
-                        production.product, production.uom,
+                        'output', production.product, production.unit,
                         production.quantity)
                     if move:
-                        move.production_output = production
                         move.unit_price = Decimal(0)
+                        move.currency = production.company.currency
                         to_save.append(move)
                 continue
 
             factor = production.bom.compute_factor(
-                production.product, production.quantity, production.uom)
+                production.product, production.quantity, production.unit)
             for input_ in production.bom.inputs:
                 quantity = input_.compute_quantity(factor)
                 product = input_.product
                 move = production._move(
-                    production.picking_location, location, company, product,
-                    input_.uom, quantity)
+                    'input', product, input_.unit, quantity)
                 if move:
-                    move.production_input = production
                     to_save.append(move)
 
             for output in production.bom.outputs:
                 quantity = output.compute_quantity(factor)
                 product = output.product
-                move = production._move(location, production.output_location,
-                    company, product, output.uom, quantity)
+                move = production._move(
+                    'output', product, output.unit, quantity)
                 if move:
-                    move.production_output = production
                     move.unit_price = Decimal(0)
+                    move.currency = production.company.currency
                     to_save.append(move)
         Move.save(to_save)
         cls._set_move_planned_date(productions)
 
-    @property
-    def _list_price_context(self):
-        return {
-            'company': self.company.id,
-            }
-
     @classmethod
     def set_cost_from_moves(cls):
         pool = Pool()
@@ -556,6 +547,7 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
         pool = Pool()
         Uom = pool.get('product.uom')
         Move = pool.get('stock.move')
+        Warning = pool.get('res.user.warning')
 
         moves = []
         for production in productions:
@@ -568,10 +560,7 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
             for input_ in production.inputs:
                 if input_.state == 'cancelled':
                     continue
-                if input_.cost_price is not None:
-                    cost_price = input_.cost_price
-                else:
-                    cost_price = input_.product.cost_price
+                cost_price = input_.get_cost_price()
                 input_quantities[input_.product] += (
                     Decimal(str(input_.internal_quantity)))
                 input_costs[input_.product] += (
@@ -586,9 +575,11 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
                     cost_price = (
                         input_costs[product] / input_quantities[product])
                     unit_price = round_price(Uom.compute_price(
-                            product.default_uom, cost_price, output.uom))
-                    if output.unit_price != unit_price:
+                            product.default_uom, cost_price, output.unit))
+                    if (output.unit_price != unit_price
+                            or output.currency != production.company.currency):
                         output.unit_price = unit_price
+                        output.currency = production.company.currency
                         moves.append(output)
                     cost -= min(
                         unit_price * Decimal(str(output.quantity)), cost)
@@ -597,11 +588,20 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
 
             for output in outputs:
                 product = output.product
-                with Transaction().set_context(production._list_price_context):
-                    list_price = product.list_price_used
+                list_price = product.list_price_used
+                if list_price is None:
+                    warning_name = Warning.format(
+                        'production_missing_list_price', [product])
+                    if Warning.check(warning_name):
+                        raise CostWarning(warning_name,
+                            gettext(
+                                'production.msg_missing_product_list_price',
+                                product=product.rec_name,
+                                production=production.rec_name))
+                    continue
                 product_price = (Decimal(str(output.quantity))
                     * Uom.compute_price(
-                        product.default_uom, list_price, output.uom))
+                        product.default_uom, list_price, output.unit))
                 prices[output] = product_price
                 sum_ += product_price
 
@@ -610,7 +610,7 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
                 for output in outputs:
                     if output.product == production.product:
                         quantity = Uom.compute_qty(
-                            output.uom, output.quantity,
+                            output.unit, output.quantity,
                             output.product.default_uom, round=False)
                         quantity = Decimal(str(quantity))
                         prices[output] = quantity
@@ -626,8 +626,10 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
                 else:
                     quantity = Decimal(str(output.quantity))
                     unit_price = round_price(cost * ratio / quantity)
-                if output.unit_price != unit_price:
+                if (output.unit_price != unit_price
+                        or output.currency != production.company.currency):
                     output.unit_price = unit_price
+                    output.currency = production.company.currency
                     moves.append(output)
         Move.save(moves)
 
@@ -750,9 +752,13 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
         Move = pool.get('stock.move')
         Date = pool.get('ir.date')
         Move.do([m for p in productions for m in p.inputs])
-        cls.write([p for p in productions if not p.effective_start_date], {
-                'effective_start_date': Date.today(),
-                })
+        for company, productions in groupby(
+                productions, key=lambda p: p.company):
+            with Transaction().set_context(company=company.id):
+                today = Date.today()
+            cls.write([p for p in productions if not p.effective_start_date], {
+                    'effective_start_date': today,
+                    })
 
     @classmethod
     @ModelView.button
@@ -764,9 +770,13 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
         Date = pool.get('ir.date')
         cls.set_cost(productions)
         Move.do([m for p in productions for m in p.outputs])
-        cls.write([p for p in productions if not p.effective_date], {
-                'effective_date': Date.today(),
-                })
+        for company, productions in groupby(
+                productions, key=lambda p: p.company):
+            with Transaction().set_context(company=company.id):
+                today = Date.today()
+            cls.write([p for p in productions if not p.effective_date], {
+                    'effective_date': today,
+                    })
 
     @classmethod
     @ModelView.button_action('production.wizard_production_assign')
@@ -778,15 +788,25 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
     def assign_try(cls, productions):
         pool = Pool()
         Move = pool.get('stock.move')
-        if Move.assign_try(
-                [m for p in productions for m in p.assign_moves]):
+        to_assign = [
+            m for p in productions for m in p.assign_moves
+            if m.assignation_required]
+        if Move.assign_try(to_assign):
             cls.assign(productions)
-            return True
         else:
-            return False
+            to_assign = []
+            for production in productions:
+                if any(
+                        m.state in {'staging', 'draft'}
+                        for m in production.assign_moves
+                        if m.assignation_required):
+                    continue
+                to_assign.append(production)
+            if to_assign:
+                cls.assign(to_assign)
 
     @classmethod
-    def _get_reschedule_domain(cls, date):
+    def _get_reschedule_planned_start_dates_domain(cls, date):
         context = Transaction().context
         return [
             ('company', '=', context.get('company')),
@@ -794,26 +814,70 @@ class Production(ShipmentAssignMixin, Workflow, ModelSQL, ModelView):
             ('planned_start_date', '<', date),
             ]
 
+    @classmethod
+    def _get_reschedule_planned_dates_domain(cls, date):
+        context = Transaction().context
+        return [
+            ('company', '=', context.get('company')),
+            ('state', '=', 'running'),
+            ('planned_date', '<', date),
+            ]
+
     @classmethod
     def reschedule(cls, date=None):
         pool = Pool()
         Date = pool.get('ir.date')
         if date is None:
             date = Date.today()
-        productions = cls.search(cls._get_reschedule_domain(date))
-        for production in productions:
+
+        to_reschedule_start_date = cls.search(
+            cls._get_reschedule_planned_start_dates_domain(date))
+        to_reschedule_planned_date = cls.search(
+            cls._get_reschedule_planned_dates_domain(date))
+
+        for production in to_reschedule_start_date:
             production.planned_start_date = date
             production.on_change_planned_start_date()
-        cls.save(productions)
+
+        for production in to_reschedule_planned_date:
+            production.planned_date = date
+
+        cls.save(to_reschedule_start_date + to_reschedule_planned_date)
 
     @property
+    @fields.depends('warehouse')
     def picking_location(self):
         if self.warehouse:
             return (self.warehouse.production_picking_location
                 or self.warehouse.storage_location)
 
     @property
+    @fields.depends('warehouse')
     def output_location(self):
         if self.warehouse:
             return (self.warehouse.production_output_location
                 or self.warehouse.storage_location)
+
+
+class Production_Lot(metaclass=PoolMeta):
+    __name__ = 'production'
+
+    @classmethod
+    @ModelView.button
+    @Workflow.transition('done')
+    def done(cls, productions):
+        pool = Pool()
+        Lot = pool.get('stock.lot')
+        Move = pool.get('stock.move')
+        lots, moves = [], []
+        for production in productions:
+            for move in production.outputs:
+                if not move.lot and move.product.lot_is_required(
+                        move.from_location, move.to_location):
+                    move.add_lot()
+                    if move.lot:
+                        lots.append(move.lot)
+                        moves.append(move)
+        Lot.save(lots)
+        Move.save(moves)
+        super().done(productions)
